/*
 *   High Score Runtime
 *
 *   2001/01/03 V1.00  by oga.
 *
 */

#define SC_USE_X 1

#ifdef SC_USE_X
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#endif /* SC_USE_X */

#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#include "score.h"

typedef struct HiEnt {
    int  score;
    char name[256];
    char date[16];
    int  stage;
} HiEnt;

/* defines */
#define MAX_HIGHS 15

/* globals */
char  gHighPath[2048];
HiEnt gHiEnt[MAX_HIGHS + 1];
int   gCurRank = 0;

#ifdef SC_USE_X
unsigned int gColor[8];
GC           gGc[8];
Font         gFont;
#endif /* SC_USE_X */

/*
 *  Set HIGH SCORE Store Path
 *
 *  IN  hipath : High Score file path
 */
void ScInit(char *hipath)
{
    strcpy(gHighPath, hipath);
    memset(&gHiEnt, 0, sizeof(gHiEnt));

}

void ScInitX(char *hipath, Display *disp, Window win)
{
    int i;
    ScInit(hipath);

    /* set color */
    gColor[0] = MyColor(disp,"Black");
    gColor[1] = MyColor(disp,"blue");
    gColor[2] = MyColor(disp,"red");
    gColor[3] = MyColor(disp,"magenta");
    gColor[4] = MyColor(disp,"green");
    gColor[5] = MyColor(disp,"cyan");
    gColor[6] = MyColor(disp,"yellow");
    gColor[7] = MyColor(disp,"white");

    for (i = 0; i<8; i++) {
        gGc[i] = XCreateGC(disp,win,0,0);           /* create gc         */
        XSetForeground(disp,gGc[i] ,gColor[i]);     /* set color to gc   */
        XSetBackground(disp,gGc[i] ,gColor[1]);     /* set bgcolor to gc */
    }

    /* load font */
    gFont = XLoadFont(disp, "7x14bold");
    XSetFont(disp,gGc[0], gFont);
    XSetFont(disp,gGc[1], gFont);
    XSetFont(disp,gGc[2], gFont);
    XSetFont(disp,gGc[3], gFont);
    XSetFont(disp,gGc[4], gFont);
    XSetFont(disp,gGc[5], gFont);
    XSetFont(disp,gGc[6], gFont);
    XSetFont(disp,gGc[7], gFont);
}

unsigned long ScGetColor(Display *disp, char *color)
{
    Colormap cmap;
    XColor   c0,c1;
    int code;

    cmap = DefaultColormap(disp,0);
    code = XAllocNamedColor(disp,cmap,color,&c1,&c0);
    if (code) {
	return(c1.pixel);
    } else {
	return(-1);
    }
}

/*
 *  Get Today String
 *
 *  OUT : buf ... YYYY/MM/DD
 */
void ScGetToday(char *buf)
{
    time_t tt;
    time(&tt);
    
    strftime(buf,16,"%Y/%m/%d",localtime(&tt));
}

/*
 *  Get Top Score & Name
 *
 *  IN  : name ... name store area
 *  OUT : high ... top score
 *  OUT : name ... top name
 */
int ScGetHighScore(int *high, char *name)
{
    *high = gHiEnt[0].score;
    if (name) {
        strcpy(name, gHiEnt[0].name);
    }
}

/*
 *   Load HIGH SCORE
 * 
 *   high score file format
 *   <SCORE>:<NAME>:<DATE>:<STAGE> 
 */
void ScLoadHighScore()
{
    FILE *fp;
    char *pt;
    char buf[4096];
    int  i;

    if (!(fp = fopen(gHighPath,"r"))) {
        if (errno == ENOENT) {
            for (i = 0; i < MAX_HIGHS; i++) {
                gHiEnt[i].score = (MAX_HIGHS-i) * 1000;
                strcpy(gHiEnt[i].name, "nobody");
                ScGetToday(gHiEnt[i].date);
                gHiEnt[i].stage = 0;
            }
            return;
        }
    	perror("fopen");
    	printf("High Score file (%s) read open error.\n",buf);
    	return;
    }
    for (i = 0; i < MAX_HIGHS; i++) {
        if (fgets(buf, sizeof(buf), fp) == NULL) {
            /* EOF */
            break;
        } 
        if (buf[strlen(buf)-1] == 0x0a) {
            buf[strlen(buf)-1] = '\0';
        }
        if (pt = (char *)strtok(buf,":"))  gHiEnt[i].score = atoi(pt);
        if (pt = (char *)strtok(NULL,":")) strcpy(gHiEnt[i].name,pt);
        if (pt = (char *)strtok(NULL,":")) strcpy(gHiEnt[i].date,pt);
        if (pt = (char *)strtok(NULL,":")) gHiEnt[i].stage = atoi(pt);
    }

    fclose(fp);
}

/*
 *   Save HIGH SCORE
 */
void ScGetMyName(char *name)
{
    char buf[4096];
    struct passwd *pwp = getpwuid(getuid());
    char *pt;

    if (pwp) {
	strcpy(buf,pwp->pw_gecos);
	pt = (char *)strchr(buf,',');
	if (pt) {
	    *pt = '\0';
	}
        sprintf(name,"%s (%s)",buf,pwp->pw_name);
    } else {
        sprintf(name,"anonymous (%d)",getuid());
    }
}

/*
 *   Save HIGH SCORE
 */
void ScSaveHighScore()
{
    FILE *fp;
    char *pt;
    int  i;

    if (!(fp = fopen(gHighPath,"w"))) {
    	perror("fopen");
    	printf("High Score file (%s) write open error.\n",gHighPath);
    	return;
    }
    for (i = 0; i < MAX_HIGHS; i++) {
        fprintf(fp,"%d:%s:%s:%d\n",
            gHiEnt[i].score,
            strlen(gHiEnt[i].name)?gHiEnt[i].name:" ",
            strlen(gHiEnt[i].date)?gHiEnt[i].date:" ",
            gHiEnt[i].stage);
    }
    fclose(fp);
}

/*
 *   Regist High Score
 *
 *   IN  : score ... score
 *   IN  : stage ... stage
 *   IN  : name  ... name  (NULL: regist from uid info)
 *   OUT : ret   ... SC_RANKOUT(0): out of rank
 *               ... SC_RANKTOP(1): Rnnk in (High Score)
 *               ... SC_RANKIN (2): Rank in
 */
int ScRegHighScore(int score, int stage, char *name)
{
    int i, j;

    for (i = 0; i < MAX_HIGHS; i++) {
        if (score >= gHiEnt[i].score) {
            for (j = MAX_HIGHS-1; j >= i; j--) {
                if (j < MAX_HIGHS-1) {
                    /* shift high score */
                    memcpy(&gHiEnt[j+1],&gHiEnt[j],sizeof(gHiEnt[j]));
                }
            }
            break;
        }
    }

    gCurRank = i;	/* Current Ranking */

    if (i < MAX_HIGHS) {
        /* rank in */
        gHiEnt[i].score = score;
        gHiEnt[i].stage = stage;
        if (name) {
            strncpy(gHiEnt[i].name, name, sizeof(gHiEnt[i].name));
            gHiEnt[i].name[30] = '\0';
        } else {
            ScGetMyName(gHiEnt[i].name);
        }
        ScGetToday(gHiEnt[i].date);

        ScSaveHighScore();
        if (i == 0) {
            return SC_RANKTOP; /* 1 */
        } else {
            return SC_RANKIN;  /* 2 */
        }
    }
    return SC_RANKOUT; /* 0 */
}

void ScDispHighScore()
{
    int i;

    printf("====  HIGH SCORES  ====\n\n");
    printf("No.    Score Name                           Date        Stage\n");
    printf("-------------------------------------------------------------\n");
    for (i = 0; i < MAX_HIGHS; i++) {
        printf("%2d. %8d %-30s %-10s  %5d\n",
                i+1,
                gHiEnt[i].score,
                gHiEnt[i].name,
                gHiEnt[i].date,
                gHiEnt[i].stage);
    }
}

void ScDispHighScoreX(Display *disp, Window win, int x0, int y0)
{
    int  i;
    char str[1024];
    int  hight = 20;

    sprintf(str, "====  HIGH SCORES  ====");
    XDrawString(disp, win, gGc[6], x0+100, y0, str, strlen(str));
    sprintf(str, "No   Score Name                      Date     Stage");
    XDrawString(disp, win, gGc[5], x0, y0+20, str, strlen(str));
#if 0
    sprintf(str, "---------------------------------------------------");
    XDrawString(disp, win, gGc[5], x0, y0+30, str, strlen(str));
#endif

    for (i = 0; i < MAX_HIGHS; i++) {
        sprintf(str, "%2d %7d %-25s %-10s %3d\n",
                i+1,
                gHiEnt[i].score,
                gHiEnt[i].name,
                gHiEnt[i].date,
                gHiEnt[i].stage);

	XDrawString(disp, win, (i == gCurRank)?gGc[3]:gGc[7], 
	            x0, hight*i+y0+40, str, strlen(str));
    }
    sprintf(str, "Press 's' to Start.");
    XDrawString(disp, win, gGc[4], x0+120, y0+(MAX_HIGHS+1)*hight+40, str, strlen(str));
    sprintf(str, "Press 'q' to Quit.");
    XDrawString(disp, win, gGc[4], x0+120, y0+(MAX_HIGHS+1)*hight+40+15, str, strlen(str));
    XFlush(disp);
}

/* #define EXECUTABLE 1 */
#ifdef EXECUTABLE
int main(int a, char *b[])
{
    int ret;

    if (a < 3) {
        printf("usage: score <score> <stage> [<name>]\n");
        return 1;
    }

    ScInit("/tmp/highscore");                            /* 1 */
    ScLoadHighScore();                                   /* 2 */

    ScDispHighScore();

    if (a >= 4) {
        /* use specified name */
        ret = ScRegHighScore(atoi(b[1]), atoi(b[2]), b[3]);  /* 3 */
    } else {
        /* get name from uid */
        ret = ScRegHighScore(atoi(b[1]), atoi(b[2]), NULL);  /* 3 */
    }
    switch (ret) {
      case SC_RANKOUT:
        printf("OUT OF RANK!!\n");
        break;
      case SC_RANKIN:
        printf("RANK IN!!\n");
        break;
      case SC_RANKTOP:
        printf("HIGH SCORE!!\n");
        break;
      default:
        printf("unknown code (%d)\n", ret);
        break;
    }

    ScDispHighScore();
    return 0;
}
#endif /* EXECUTABLE */


