/*
 *     xalien.c           V0.10  97.01.26    by Hyper Halx.f oga.
 *                        V1.00  97.02.05    by Hyper Halx.f oga.
 *                        V1.01  99.02.01    fix hiname bug by oga.
 *                        V1.02  99.05.01    support joystick (for Linux)
 *                        V1.03  00.01.26    some message change
 *                        V1.04  01.01.03    support high score 15
 *
 *   CFLAGS : USE_JOYSTICK : use joystick for Linux
 *
 *   Copyright(C)1997,2000, Moritaka Ogasawara. 
 *   oga@mxg.mesh.ne.jp
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#ifdef USE_JOYSTICK
#include <linux/joystick.h>
#endif /* USE_JOYSTICK */

#include "xalien.h"
#include "usleep.h"
#include "score.h"

#define VER     	"1.04"
#define WAIT    	80000	/* 80000 (0.08sec) */
#define GOSTRAIGHT	2
#define usleep		u_sleep

/* for put_maze */
#define PUTBG		1
#define PUTWALL		2
#define PUTKEV		4

#ifdef USE_JOYSTICK
#define	JOY_HELP	"[-joy {0|1}] "
#else  /* USE_JOYSTICK */
#define	JOY_HELP	""
#endif /* USE_JOYSTICK */

unsigned long MyColor();

int X_SIZE   = CSIZ*23;
int Y_SIZE   = CSIZ*31 + 4;
int MAZENO   = 0;
int PLAYER   = 1;
int CREDIT   = 1;
int SERVER   = 0;		/* invoke with -s     */
int HI_SCORE = 5000;
int numAL;			/* ALIEN̐          */
int restAL;			/* ALIEN̎c        */
int high;			/* Hi-ScoreBtO */
int xwait;
int over  = 0;			/* game over flag     */
int stage = -1;                 /* Stage no - 1       */
int joyf     = 0;		/* 1: Use Joystick       V1.02 */
int joyinit  = 0;		/* get_joystick init flg V1.02 */
int joy_x_max= 0;		/* joy maxvalue x        V1.02 */
int joy_y_max= 0;		/* joy maxvalue y        V1.02 */
char HI_NAME[256];
char joydev[256] = "/dev/null"; /* joystick device name  V1.02 */

Display *d;
Window  w;
GC      gc[COLORS];

extern	Pixmap	pix[];

XEvent	report, e;

char	map[46*54];			/* maze map in memory    */
char	diff[5]  = {-46, 92, -1, 2, 0};	/* diff  of move (2step) */
char	diff2[5] = {-45, 93, 45,48, 0};	/* diff2 of move (2step) */
al_t	al[32];				/* alien data            */
kv_t	kv[2];				/* kebiishi data         */
ana_t	ana[16];			/* hole data             */
enum	{ALL, ONLYONE};			/* for init_alien()      */

int main(a,b)
int a;
char *b[];
{
	XSetWindowAttributes att;
	Font	font, fontbig;

	unsigned long color[COLORS];
	int i, j, c;
	int x1, x2, x3, x4;
	int y1, y2, y3, y4;
	int level;
	int gcnt;			/* global counter       */
	int tmoutf;			/* timeout flag         */
	int timeout;			/* timeout cnt 500:1min */
	int clearf = 0;                 /* hi-score disp clear  */
	char name[100];			/* window title name    */
	char fontn[256];		/* normal font name     */
	char fontbn[256];		/* big font name        */
	char hostn[32];			/* server hostname      */
	char highpath[1024];		/* HiScore file path    */
	char highname[1024];		/* HiScore Name         */
	struct timeval tv;

	/* get args */
	for (i = 1; i < a; i++) {
	    if (!strncmp(b[i],"-h",2)) {
		printf("XALIEN  Ver %s  (C)1997,2001, Moritaka Ogasawara.\n",VER);
		printf("                  (C)1980, DENKI ONKYO for Arcade\n",VER);
		printf("                  (C)1979, Tokyo Univ. TSG for 8bit PC\n\n",VER);
#ifdef NET_SUPPORT
		printf("usage : xalien %s[{-s | <hostname>}] [-name <your_name>] [-h]\n\n", JOY_HELP);
		printf("                -s : invoke as server(1st player)\n");
		printf("        <hostname> : invoke as client(2nd player)\n");
#else
		printf("usage : xalien %s[-name <your_name>] [-h]\n\n",JOY_HELP);
#endif
		exit(1);
	    }
	    if (!strncmp(b[i],"-s",2)) {
	        PLAYER = 2;
	        SERVER = 1;
	        continue;
	    }
	    if (i+1 < a && !strncmp(b[i],"-joy",4)) {
	        /* use joystick */
	        sprintf(joydev, "/dev/js%s", b[++i]);
	        joyf = 1;
	        continue;
	    }
	    if (i+1 < a && !strcmp(b[i],"-name")) {  /* V1.04-A */
	        strcpy(highname, b[++i]);
	        continue;
	    }
            strcpy(hostn,b[i]);
	}

	d = XOpenDisplay(NULL);

	/* set default font name */
#ifdef LINUX
#define FONTSP		(CSIZ-12)
	strcpy(fontn, "-*-helvetica-*-r-*-*-12-*-*-*-*-*-*-*");
	strcpy(fontbn,"-*-helvetica-*-o-*-*-24-*-*-*-*-*-*-*");
#else
#define FONTSP		(CSIZ-13)
	strcpy(fontn,"7x13bold");
	strcpy(fontbn,"-*-helvetica-*-o-*-*-24-*-*-*-*-*-*-*");
#endif

	w = XCreateSimpleWindow(d, RootWindow (d,0),
					0, 0 ,          /* home x,y         */ 
					X_SIZE,Y_SIZE,  /* window size x,y  */
					5,              /* border_width     */
					1,              /* border           */
					0   );          /* background       */

	att.override_redirect = 0;         /* Window Manager  ... 0 */
                                           /*                    Ȃ ... 1 */

	XChangeWindowAttributes(d,w,CWOverrideRedirect,&att);
	XMapWindow(d,w);

	/* put window title */
	sprintf(name,"XALIEN Ver%s",VER);
	XStoreName(d,w,name);

	/* set color */
	color[BLACK]   = MyColor(d,"Black");
	color[BLUE]    = MyColor(d,"navy blue");
	color[RED]     = MyColor(d,"red");
	color[MAGENTA] = MyColor(d,"magenta");		/* not use */
#ifdef LINUX
	color[GREEN]   = MyColor(d,"light green");
#else
	color[GREEN]   = MyColor(d,"#80ee80");
#endif
	color[CYAN]    = MyColor(d,"cyan");
	color[YELLOW]  = MyColor(d,"yellow");
	color[WHITE]   = MyColor(d,"white");
	for (i = COL8; i <COLORS; i++) {
	    color[i]   = MyColor(d,"white");
	}
	color[COL8]    = MyColor(d,"#ff9933");		/* orange     */
	color[COL9]    = MyColor(d,"#007700");		/* dark green */

	for (i = 0; i<COLORS; i++) {
	    gc[i] = XCreateGC(d,w,0,0);            /* create gc         */
	    XSetForeground(d,gc[i] ,color[i]);     /* set color to gc   */
	    XSetBackground(d,gc[i] ,color[BLUE]);  /* set bgcolor to gc */
	}

	/* load font */
	font    = XLoadFont(d, fontn);
	fontbig = XLoadFont(d, fontbn);
	XSetFont(d,gc[BLUE],   font);
	XSetFont(d,gc[RED],    font);
	XSetFont(d,gc[MAGENTA],fontbig);
	XSetFont(d,gc[GREEN],  font);
	XSetFont(d,gc[CYAN],   font);
	XSetFont(d,gc[YELLOW], font);
	XSetFont(d,gc[WHITE],  font);
	XSetFont(d,gc[COL8],   fontbig);
	XSetFont(d,gc[COL9],   font);

	/* pixmap initialize */
	pix_init();

	XSelectInput(d,w, ButtonPressMask | ExposureMask |
			  KeyPressMask | KeyReleaseMask);
	XFlush(d);		/* dummy */
	sleep(1);

	/* initialize random seed */
	gettimeofday(&tv,0);
	srand(tv.tv_usec);

        /* Init & Load High Score  V1.04-A start*/
        sprintf(highpath,"%s/xalien.high",HIGHPATH);
        /* ScInit(highpath); */
        ScInitX(highpath, d, w);
	ScLoadHighScore();
	ScGetHighScore(&HI_SCORE, HI_NAME);
        /* ScDispHighScore(); */
        ScDispHighScoreX(d, w, 5, 60);
        PUT_LOGO(90,3,RED);

        while(1) {
            if (get_key() == START) break;
            usleep(100000);
        }
	/* V1.04-A end */

start:
	stage = 0;
	over  = 0;
	high  = 0;
	clearf= 0;
	memset(kv, 0, sizeof(kv));
	kv[0].rest  = 3;		/* bug fix V1.02 */
	kv[1].rest  = 0;

	/* stage loop */
	while (1) {
	  /* stage initialize */

	  gcnt   = 0;
	  tmoutf = 0;
	  MAZENO = rnd(7);
	  make_map();				/* make maze map on mem  */
	  restAL = numAL = (stage % 3)*2+4;
	  put_maze(PUTBG+PUTWALL+PUTKEV);	/* put maze on screen    */
	  XFlush(d);
	  init_alien(numAL,0);			/* Initialize all ALIEN  */
	  init_kevi(PLAYER);			/* Initialize KEBIISHI   */

	  level = (stage % 9) / 3;             /* 0 - 2 */
	  if (level == 2) {
	      xwait   = WAIT/2;
	      timeout = 2500*2;
	  } else {
	      xwait   = WAIT;
	      timeout = 2500;
	  }

	  /* game loop */
	  while (1) {
	    ++gcnt;
	    if (gcnt > 2500 && (gcnt % 10 == 0)) {
	        /* Ԑ؂ (gcnt=500/1min */
	        if (!tmoutf) {
	            printf("");
	            fflush(stdout);
	            printf("");
	            fflush(stdout);
	            tmoutf = 1;
	        }
	        AddOneAlien();
	    }

	    /* gړ */
	    if (move_kebi()) {
	    	/* Go To Next Stage!! */
	    	sleep(2);
	    	break;
	    }

	    /* ALIEN ړ */
	    if (level == 0) {
	        if ((gcnt & 1) && move_alien()) {
	            printf("");
	            fflush(stdout);
	    	    over = 1;
		    if (high) {			/* V1.01-C         */
		        /* SaveHighScore();        V1.04-D         */
		        high = 0;		/* V1.01-A         */
		    }
		    ScRegHighScore(kv[0].score, stage+1, 
		               strlen(highname)?highname:NULL); /* V1.04-A */
		    ScGetHighScore(&HI_SCORE, HI_NAME);         /* V1.04-A */
		    /* ScDispHighScore(); for Text */
		    put_maze(PUTWALL);		/* update HI-SCORE */
	    	}
	    } else {
	        if (move_alien()) {
	            printf("");
	            fflush(stdout);
	    	    over = 1;
		    if (high) {			/* V1.01-C         */
		        /* SaveHighScore();        V1.04-D         */
		        high = 0;		/* V1.01-A         */
		    }
		    ScRegHighScore(kv[0].score, stage+1,
		               strlen(highname)?highname:NULL); /* V1.04-A */
		    ScGetHighScore(&HI_SCORE, HI_NAME);         /* V1.04-A */
		    /* ScDispHighScore(); for Text */
		    put_maze(PUTWALL);		/* update HI-SCORE */
	    	}
	    }
	    if (over) {
	        if (get_key() == START) goto start; 
	        XDrawString(d, w, gc[COL8], X_SIZE/2-80, Y_SIZE/2+8,
				"GAME  OVER", strlen("GAME  OVER"));
		if ((gcnt % 200) < 100) {
		    clearf = 0;
		    ScDispHighScoreX(d, w, 5, 60); /* V1.04-A */
		} else {
		    if (clearf == 0) {
		        /* clear High Score */
		        put_maze(PUTBG+PUTWALL);
		    }
		    clearf = 1;
		}
		XFlush(d);
	    }

	    if (XCheckMaskEvent(d,ButtonPressMask,&report) == True) {
	    	dump_map();
	    }
	    if (XCheckMaskEvent(d,ExposureMask,&report) == True) {
		Window root;
		int x,y;
		unsigned int width,height, border, depth;

		XGetGeometry(d,w,&root,&x,&y,&width,&height,&border,&depth);
		if (X_SIZE != width || Y_SIZE != height) {
		    X_SIZE = width;
		    Y_SIZE = height;
		    XClearWindow(d,w);
		}
	    }
	    usleep(xwait);
          }	/* game loop */
          stage += 1;
	} /* stage loop */
	return 0;
}

unsigned long MyColor(display,color)
Display *display;
char *color;
{
	Colormap cmap;
	XColor c0,c1;
	int code;

	cmap = DefaultColormap(display,0);

	code = XAllocNamedColor(display,cmap,color,&c1,&c0);
	if (code)
		return(c1.pixel);
	else
		return(-1);
}

/* 
 *  put maze on screen 
 *
 *  IN level : PUTBG   ... fill background
 *             PUTWALL ... put wall 
 *             PUTKEV  ... put kebiishi to map (for start of stage)
 */
void put_maze(level)
int level;
{
	int i, j, c, base;
	char buf[128];
	int  wall_color = GREEN;

	if (over) {
	    wall_color = COL9;
	}

	base = CSIZ*2;

	if (level & PUTBG) {
	    /* bgcolor */
	    XFillRectangle(d,w,gc[BLUE] ,0, 0, X_SIZE, Y_SIZE);
	}

	if (level & PUTWALL) {
	    /* O */
	    XFillRectangle(d,w,gc[wall_color],0, base, CSIZ*23, CSIZ/2);
	    XFillRectangle(d,w,gc[wall_color],0, base, CSIZ/2,  CSIZ*27);
	    XFillRectangle(d,w,gc[wall_color],CSIZ*23-CSIZ/2, base, 
						CSIZ/2,  CSIZ*27);
	    XFillRectangle(d,w,gc[wall_color],0, base+CSIZ*27-CSIZ/2, 
						CSIZ*23,  CSIZ/2);
	    /*  */
	    for (j = 0; j<11; j++) {
	        for (i = 0; i<9; i++) {
	    	    if (maze[MAZENO][i+j*9]) {
		        XFillRectangle(d,w,gc[wall_color],
		    		40+i*CSIZ*2, base+40+j*CSIZ*2,
		    		CSIZ*2, CSIZ*2);
	    	    }
	        }
    	    }
    	}

	/* \ (redrawp) */
        for (j = 0 ; j<46-4; j += 2) {
	    for (i = 0; i<54-4; i += 2) {
		c = map[BASE+ 46*i +j];		/* V1.02 */
		if (ANA1 <= c && c <= UME4) {
#ifdef DEBUG
		    printf("### PUT_CHAR(%d,%d,%d)\n",i,j,c);
#endif
		    PUT_CHAR(j,i,c,WHITE);	/* V1.02 */
		}
	    }
	}

	/* SCORE \ */
	XDrawString(d, w, gc[WHITE], 1*CSIZ, CSIZ-FONTSP,
				"SCORE", strlen("SCORE"));
	XDrawString(d, w, gc[WHITE], 9*CSIZ, CSIZ-FONTSP, 
				"HI-SCORE", strlen("HI-SCORE"));
	if (level & PUTKEV) {
	    map[BASE+46*48+20] = KEVA3;		/* g u */
        }
    	if (PLAYER > 1) {
	    XDrawString(d, w, gc[WHITE], 1*CSIZ, CSIZ-FONTSP,
				"SCORE<1>", strlen("SCORE<1>"));
	    XDrawString(d, w, gc[WHITE], 18*CSIZ, CSIZ-FONTSP,
				"SCORE<2>", strlen("SCORE<2>"));
	    if (level & PUTKEV) {
	        map[BASE+46* 0+20] = KEVA3;	/* g u */
    	    }
	}

	/* HI-SCORE name */
	XFillRectangle(d,w,gc[BLUE], 1*CSIZ, base+CSIZ*27, Y_SIZE, CSIZ);
	sprintf(buf,"HI-SCORE : %s ",HI_NAME);
	XDrawString(d, w, gc[WHITE], 1*CSIZ, base+CSIZ*28-FONTSP, 
							buf, strlen(buf));
	score_disp();

	/* STAGE V1.04-A */
	sprintf(buf,"Stage %d",stage+1);
	XDrawString(d, w, gc[WHITE], 19*CSIZ, base+CSIZ*28-FONTSP, 
							buf, strlen(buf));
	/* CREDITS */
	sprintf(buf,"CREDIT %d",CREDIT);
	XDrawString(d, w, gc[WHITE], 19*CSIZ, base+CSIZ*29-FONTSP, 
							buf, strlen(buf));
}

/* 
 *   map̋`codeŖ߂ 
 */
void map_fill_rect(x,y,w,h,code)
int x,y,w,h,code;
{
	int i,j;

#ifdef DEBUG
	if (x+w >46 ) printf("map_fill_rect: width over!  x=%d w=%d\n",x,w);
	if (y+h >54 ) printf("map_fill_rect: height over! y=%d h=%d\n",y,h);
#endif
	for (j = y; j < y+h; j++) {
	    for (i = x; i < x+w; i++) {
		map[j*46+i] = code;
	    }
	}
}

/* mapɕǂ */
void make_map()
{
	int i, j, base;
	char buf[128];

	memset(map, 0, sizeof(map));
	/* Og */
	map_fill_rect( 0, 0,  46, 2, KABE);
	map_fill_rect( 0,52,  46, 2, KABE);
	map_fill_rect( 0, 0,   2,54, KABE);
	map_fill_rect(44, 0,   2,54, KABE);

	for (j = 0; j<11; j++) {
	    for (i = 0; i<9; i++) {
	    	if (maze[MAZENO][i+j*9]) {
		    map_fill_rect(i*4+4, j*4+4, 6, 6, KABE);
	    	}
	    }
    	}
}

/* GENERATE RANDOM VALUE */
int rnd(n)
int n;
{
#ifdef LINUX
	return((int)(((float)rand()*n)/RAND_MAX));
#else
	return(rand()*n/RAND_MAX);
#endif
}

/*
 *   move ALIEN 
 * 
 *   OUT  ret: 0 ... game continue
 *             1 ... game is over!!
 */
int move_alien()
{
    int i, j, c;
    struct timeval tv,tv2;

    for (i = 0; i < numAL; i++) {
	if (al[i].die) {
	    /* łskip */
	    continue;
	}

	if (al[i].droptv.tv_sec) {
	    /* ɗĂ */
	    ++al[i].dropcnt;
	    if (!al[i].stop) {
                c = GET_CHAR(al[i].x, al[i].y);
                if (ZUBO0 <= c && c <= ZUBO1 ) {
	    	    PUT_CHAR(al[i].x, al[i].y, (al[i].dropcnt & 1) + ZUBO0, 
	    	    							WHITE);
	    	}
	    }
	    gettimeofday(&tv,0);
	    tv_diff(&tv, &al[i].droptv, &tv2);
	    if ((al[i].stop && tv2.tv_sec>0 && tv2.tv_usec>(al[i].ana-ANA1)*200000)
	          || (!al[i].stop && tv2.tv_sec > 5)) {
	        /* 甇o */
	        memset(&al[i].droptv, 0, sizeof(struct timeval));
	        al[i].stop = 0;
	        al[i].ana  = 0;
	    } else {
	        continue;
	    }
	}

	/* _̃`FbN */
	if (((al[i].x) % 8 == 0) && ((al[i].y) % 8 == 0)) {
	    if (al[i].cnt) {
		--al[i].cnt;
	    } else {
		al[i].cnt = rnd(4);
		al[i].vec = get_newvec(al[i].x, al[i].y);
	    }
	}
	/* s`FbN */
	c = get_next_chr(al[i].x, al[i].y, al[i].vec);
	if (ALIEN0 <= c && c <= KABE) {
	    /* ǂȂVݒ */
	    al[i].cnt = rnd(GOSTRAIGHT);
	    al[i].vec = get_newvec(al[i].x, al[i].y);
	} else if (ANA1 <= c && c <= ZUBO1) {
	    /* ꍇ邩 */
	    int aa, xx, yy;

	    xx = al[i].x;
	    yy = al[i].y;
	    update_xy(&xx, &yy, al[i].vec);
	    update_xy(&xx, &yy, al[i].vec);
#ifdef DEBUG
	    printf("\n %d,%d / %d\n",al[i].x,al[i].y,al[i].vec);
	    for (aa = 0; aa<numAL; aa++) {
	    	if (al[aa].droptv.tv_sec) {
	    	    printf("DROPING X=%d Y=%d\n",al[aa].x,al[aa].y);
	    	}
	    }
#endif
	    aa = SearchAlien(xx,yy);

	    /* 邩 */
	    if (aa >= 0) {
	        /*  */
	        al[aa].stop = 0;
	        al[aa].ana  = 0;
	        memset(&al[aa].droptv, 0, sizeof(struct timeval));
	        PUT_ALIEN(al[aa].x, al[aa].y, al[aa].vec);

	        al[i].cnt = rnd(GOSTRAIGHT);
	        al[i].vec = get_newvec(al[i].x, al[i].y);
	    } else if (IS_BLOCK(al[i]) && ANA1 <= c && c <= ANA5) {
	        /*  */
	        gettimeofday(&al[i].droptv,0);
	        if (c != ANA5) {
	            /* sSȌ */
	            al[i].ana  = c;  /* p^[ۑ */
	            al[i].stop = 1;
	        }
	    }
	} else if (KEVA0 <= c && c <= KEVC3) {
	    /* gς */
	    int xx, yy, die[2];

	    die[0] = die[1] = 0;

	    /* s */
	    xx = al[i].x;
	    yy = al[i].y;
	    update_xy(&xx, &yy, al[i].vec);
	    update_xy(&xx, &yy, al[i].vec);
	    CLEAR_CHAR(xx, yy);

	    /* g &  */
    	    for (j = 0; j<PLAYER; j++) {
                if (map[BASE+kv[j].x+kv[j].y*46]==SPACE       ||
                       map[BASE+kv[j].x+kv[j].y*46+1]==SPACE  ||	
                       map[BASE+kv[j].x+kv[j].y*46+46]==SPACE ||	
                       map[BASE+kv[j].x+kv[j].y*46+47]==SPACE ) {
            	    CLEAR_CHAR(kv[j].x, kv[j].y);
            	    die[j] = 1;
                }
    	    }
	    PUT_PAKU(al[i].x, al[i].y, al[i].vec);
	    sleep(2);

	    /* ςL */
	    CLEAR_CHAR(xx, yy);
	    PUT_ALIEN(al[i].x, al[i].y, al[i].vec);	/* ALIENĕ\ */

            /* S & ̌g\ */
    	    for (j = 0; j<PLAYER; j++) {
                if (kv[j].rest > 0 && die[j]) {
            	    kev_die_newpos(j);
                }
    	    }
    	    if ((kv[0].rest + kv[1].rest) == 0) {
    	        return 1;	/* game is over */
    	    }
	} 
	if (al[i].vec != NOMOVE) {	
	    /* ړ̏ꍇ̂ݕ\ */
	    CLEAR_CHAR(al[i].x, al[i].y);
	    update_xy(&al[i].x, &al[i].y, al[i].vec);	/* ̏ꏊɈړ */
	    if (al[i].stop) {
	        /* @肩̌ɗĂꍇ͌̃p^[\ */
	        PUT_CHAR(al[i].x, al[i].y, al[i].ana, WHITE);
	    } else {
	        if (al[i].droptv.tv_sec) {
	            /* ɗꍇ */
	    	    PUT_CHAR(al[i].x, al[i].y, (al[i].dropcnt & 1) + ZUBO0, 
	    	    							WHITE);
	    	} else {
	            PUT_ALIEN(al[i].x, al[i].y, al[i].vec);
	        }
	    }
	}

    }
    return 0;
}

/*
 *   ALIEN Cǉ (Ԑ؂ꎞ)
 */
void AddOneAlien()
{
    int i;

    if (numAL < 32) {
    	init_alien(numAL,ONLYONE);
    	++numAL;
    	++restAL;
    } else {
        for (i=0; i<32; i++) {
            if (al[i].die) {
                init_alien(i,ONLYONE);
                ++restAL;
            }
        }
    }
}

/* 
 *   gVʒuɕ\
 *
 */
void kev_die_newpos(num)
int num;
{
    int i;

#ifdef DEBUG
    if (kv[num].rest == 0) {
        printf("kev_die_newpos: player %d maybe zonbie!?\n",num+1); 
    }
#endif
    --kv[num].rest;
    CLEAR_CHAR(kv[num].x, kv[num].y);
    kev_rest_disp();				/* g̎c\  */
    if (kv[num].rest > 0) {
	get_rnd_pos(&kv[num].x, &kv[num].y);
	kv[num].vec     = RIGHT;		/* right direction     */

    	for (i = 0; i < 4; i++) {
            CLEAR_CHAR(kv[num].x, kv[num].y);
	    XFlush(d);
            usleep(150000);			/* wait 0.15sec */
	    PUT_KEV(kv[num].x, kv[num].y, kv[num].vec);
	    XFlush(d);
            usleep(150000);			/* wait 0.15sec */
    	}
    }
}

/* 
 *   g̎c\
 *
 */
void kev_rest_disp()
{
    int i,j;

    for (i=0; i<PLAYER; i++) {
    	for (j=0; j<3; j++) {
    	    if (kv[i].rest > j+1) {
		XCopyPlane(d,pix[KEVA3],w,gc[YELLOW], 0, 0, CSIZ, CSIZ,
				X2X(i*28+j*2), Y2Y(54), 1);
    	    } else {
		XFillRectangle(d,w,gc[BLUE] ,
				X2X(i*28+j*2), Y2Y(54), CSIZ, CSIZ);
    	    }
	}
    }
    XFlush(d);
}

/* 
 *   XRA\
 *
 */
void score_disp()
{
    int  i,j;
    char buf[32];

    for (i=0; i<PLAYER; i++) {
        sprintf(buf,"%6d",kv[i].score);
	XFillRectangle(d,w,gc[BLUE] ,
				(1+i*17)*CSIZ, CSIZ, CSIZ*4, CSIZ);
	XDrawString(d,w,gc[CYAN],(1+i*17)*CSIZ,2*CSIZ-FONTSP,buf,strlen(buf));
    }

    sprintf(buf,"%6d",HI_SCORE);
    XFillRectangle(d,w,gc[BLUE] ,
				10*CSIZ, 1*CSIZ, CSIZ*4, CSIZ);
    XDrawString(d, w, gc[CYAN], 10*CSIZ, 2*CSIZ-FONTSP, buf, strlen(buf));
    XFlush(d);
}


/* 
 *   move KEBIISHI
 *
 *   OUT  ret: 0 ... game continue
 *             1 ... next stage!
 */
int move_kebi()
{
    int i, j, key, c;
    int xx, yy;
    struct timeval  tv, tv2;

    for (i = 0; i<PLAYER; i++) {
        if (!kv[i].rest) {
            /* cȂ */
            continue;
        }
        if (kv[i].digchar) { /* V1.02 */
            PUT_CHAR(kv[i].x, kv[i].y, kv[i].vec+KEVA0,YELLOW);
            kv[i].digchar = 0;
        }
        if ((kv[i].x & 1) || (kv[i].y & 1 )) {
            key = kv[i].vec;
        } else {
            /* L[XL */
	    if (i == 0) {
	        key = get_key();
	    } else {
	        /* lbg[NoRL[ */
	        /* T|[g */
	        key = 4;
#ifdef DEBUG
	        if (!kv[0].rest) {
  	            key = get_key();
	        }
#endif
	    }
        }
	if (key < NOMOVE ) {
	    kv[i].vec = key;

	    /* s`FbN */
	    c = get_next_chr(kv[i].x, kv[i].y, kv[i].vec);
	    if (c == SPACE) {
                /* gړ */
	        CLEAR_CHAR(kv[i].x, kv[i].y);
	        update_xy(&kv[i].x, &kv[i].y, kv[i].vec);  /* ̏ꏊɈړ */
	    }
            PUT_KEV(kv[i].x, kv[i].y, kv[i].vec);

	} else if (key == DIG || key == DUG) {
            /* @/ */
            gettimeofday(&tv,0);
            tv_diff(&tv, &kv[i].digtv, &tv2);
            if (tv2.tv_sec > 0 || tv2.tv_usec > 400000) {
                kv[i].digtv.tv_sec = tv.tv_sec;
                kv[i].digtv.tv_usec = tv.tv_usec;
	        xx = kv[i].x;
	        yy = kv[i].y;
	        update_xy(&xx, &yy, kv[i].vec);
	        update_xy(&xx, &yy, kv[i].vec);
	        c = GET_CHAR(xx,yy);
	        if (key == DIG) {
	            /* @   */
	            if (c == SPACE) {
	                c = ANA1;
	                PUT_CHAR(xx,yy,c,WHITE);
                        kv[i].digchar = 1;    /* V1.02 */
	            } else if (ANA1 <= c && c <= ANA4) {
	                ++c;
	                PUT_CHAR(xx,yy,c,WHITE);
                        kv[i].digchar = 1;    /* V1.02 */
	            }
                    if (kv[i].digchar) {
                        PUT_CHAR(kv[i].x, kv[i].y, kv[i].vec+KEVC0,YELLOW); /* V1.02 */
		    }
	        } else {
	            /* ߂ */
	            if (ANA1 <= c && c <= ZUBO1) {
                        kv[i].digchar = 1;    /* V1.02 */
                        PUT_CHAR(kv[i].x, kv[i].y, kv[i].vec+KEVC0,YELLOW); /* V1.02 */
	            }

	            if (ANA2 <= c && c <= ANA5) {
	                --c;
	            } else if (c == ANA1) {
	                c = SPACE;
	            } else if (c == ZUBO0 || c == ZUBO1) {
			c = UME1;
	            } else if (UME1 <= c && c <= UME3) {
	                ++c;
	            } else if (c == UME4) {
	                /* ߊ */
	            	char buf[16];
	            	int  sc;

	                CLEAR_CHAR(xx,yy);
			c = SPACE;
			j = SearchAlien(xx,yy);
			sc = TvToScore(&al[j].droptv);
			kv[i].score += sc;
			if (kv[i].score > HI_SCORE) {
			   /* update HI-SCORE */
			   HI_SCORE = kv[i].score;
			   high     = 1;
			}
			if (!kv[i].upf && kv[i].score > 5000) {
			    /* Extra 1UP */
			    ++kv[i].rest;
			    kv[i].upf = 1;
			    kev_rest_disp();
			}
			sprintf(buf,"%d",sc);
			XDrawString(d, w, gc[(sc == 1000) ? YELLOW:CYAN], 
			     X2X(xx), Y2Y(yy)+CSIZ/2, buf, strlen(buf));
		        XFlush(d);
		        sleep(1);
		        score_disp();
			XFillRectangle(d,w,gc[BLUE] ,
			     X2X(xx), Y2Y(yy)-CSIZ/2, CSIZ*2, CSIZ);
		        al[j].die = 1;
			put_maze(PUTWALL);	      /* redraw wall     */
			for (j = 0; j<PLAYER; j++) {  /* redraw kebiishi */
			    PUT_KEV(kv[j].x, kv[j].y, kv[j].vec);
			}
		        if (--restAL == 0) {
		            /* stage clear */
		            return 1;
		        }
	            }
	            if (c == SPACE) {
	                CLEAR_CHAR(xx,yy);
	            } else {
	                if (ANA1 <= c && c <= UME4) {
	                    PUT_CHAR(xx,yy,c,WHITE);
	                }
	            }
	        } /* DUG */
            } /* 0.4sec */
	} else if (key != NOMOVE) {
	    printf("move_kebi: invalid key code!! (%d)\n",key);
	}
    }
    return 0;
}

/*
 *  󂢂Ă{
 *  
 *  RET : 󂢂Ă(_)
 *        󂫂Ȃꍇ́ANOMOVE
 */
int get_newvec(x, y)
int x,y;
{
    int   cnt = 0, c, i;
    char  vecs[4];
    
    for (i = 0; i<4; i++) {
	c = get_next_chr(x,y,i);
	if (c < ALIEN0) {
	    /* s */
	    vecs[cnt] = i;
	    cnt++;
	}
    }
    if (cnt) {
	return vecs[rnd(cnt)];
    } else {
	return NOMOVE;
    }
}

/*
 *   ̍s̃LN^R[h𓾂
 *
 *   IN : vec 0-3:s  4:݂̈ʒu
 */
int get_next_chr(x,y,vec)
int x,y,vec;
{
	int c1,c2;
	c1 = map[x+y*46+diff[vec]+BASE];
	c2 = map[x+y*46+diff2[vec]+BASE];
	if (c1 == SPACE || c2 == SPACE) {
	    /* Ƃ󔒂̏ꍇ̂SPACEԂ */
	    c1 = c1+c2;
	}
	return c1;
}

/*
 *   ALIENʒu
 *
 *   IN  num : ALIEN̐ or ALIENԍ
 *      flag :     ALL: num̐       (>1)
 *             ONLYONE: ԍnumALIEN  (>0)
 */
void init_alien(num,flag) 
int num,flag;
{
    int i, start=0, end=num;

    if (flag == ONLYONE) {
        start = num;
        end   = num+1;
    }

    for (i=start; i<end; i++) {
	if (flag == ALL) usleep(600000);	/* wait 0.6sec */
	get_rnd_pos(&al[i].x, &al[i].y);
	al[i].vec     = get_newvec(al[i].x, al[i].y);
	al[i].die     = 0;
	al[i].cnt     = rnd(GOSTRAIGHT);
	al[i].stop    = 0;
	al[i].ana     = 0;
	memset(&al[i].droptv, 0, sizeof(struct timeval));
	PUT_ALIEN(al[i].x, al[i].y, al[i].vec);
	XFlush(d);
    }
    if (flag == ALL) usleep(600000);		/* wait 0.6sec */
}

/*
 *   g̈ʒu
 *
 *   IN  num : number of player
 */
void init_kevi(num) 
{
    int i, j;

    for (i=0; i<num; i++) {
	kv[i].x       = 20;
	kv[i].y       = (1-i)*48;
	kv[i].vec     = RIGHT;		/* right direction  */
	gettimeofday(&kv[i].digtv,0);
    }
    kev_rest_disp();			/* gc\ */
    for (j = 0; j < 4; j++) {
        for (i = 0; i < num; i++) {
            CLEAR_CHAR(kv[i].x, kv[i].y);
        }
	XFlush(d);
        usleep(150000);		/* wait 0.15sec */
        for (i = 0; i < num; i++) {
	    PUT_KEV(kv[i].x, kv[i].y, kv[i].vec);
        }
	XFlush(d);
        usleep(150000);		/* wait 0.15sec */
    }
}

/*
 *  NȂ_𓾂
 * 
 */
void get_rnd_pos(xp,yp) 
int *xp,*yp;
{
    do {
	*xp = rnd(6)*8;
	*yp = rnd(7)*8;
    } while (get_next_chr(*xp, *yp, NOMOVE));
}

/*
 *    isɐi
 */
void update_xy(xp,yp,vec) 
int *xp,*yp;
{
    switch (vec) {
      case UP:
	*yp -= 1;
	break;

      case DOWN:
	*yp += 1;
	break;

      case LEFT:
	*xp -= 1;
	break;

      case RIGHT:
	*xp += 1;
	break;

      case NOMOVE:
	break;

      default:
	printf("update_xy: Invalid vector! (%d)\n",vec);
	break;
    }
}

/* 
 *  x,yʒuALIENT
 *
 *  OUT ret : ALIENԍ
 */
int SearchAlien(x,y)
int x, y;
{
    int i;
    for (i = 0; i < numAL; i++) {
    	if (al[i].x == x && al[i].y == y && !al[i].die) {
    	    return i;
    	}
    }
#ifdef DEBUG
    printf("SearchAlien: Alien Not Found! (x=%d, y=%d)\n",x,y);
#endif
    return -1;
}

/* 
 *  tvɑΉXRAԂ
 *
 *  IN  tv : Alien drop time
 *  OUT ret : XRA
 */
int TvToScore(tv)
struct timeval *tv;
{
    struct timeval tv0, tv2;
    int ret = 0;

    if (rnd(10) == 1) {
    	return 1000;		/* BONUS SCORE */
    }
    gettimeofday(&tv0, 0);
    tv_diff(&tv0, tv, &tv2);
    if (tv2.tv_sec < 1) {
        ret = 400 + (numAL-4)/2 * 200;
    } else if (tv2.tv_sec < 2) {
        ret = 300 + (numAL-4)/2 * 200;
    } else if (tv2.tv_sec < 4) {
        ret = 200 + (numAL-4)/2 * 200;
    } else {
        ret = 100 + (numAL-4)/2 * 200;
    }
#ifdef DEBUG
    printf("TvToScore: spend time = %d.%06d / SCORE = %d\n",
			    			tv2.tv_sec, tv2.tv_usec, ret);
#endif
    return ret;
}

/*
 *   JOYSTICK荞  V1.02
 *
 *   IN  : Ȃ
 *   IN  : global : joydev : joystickfoCX
 *   OUT : ret 
 *           NOMOVE : L[͂Ȃ
 *           UP,DOWN,LEFT,RIGHT,DIG,DUG,START : eL[ꂽB
 *
 */
int get_joystick()
{
    int status = NOMOVE;
    int st;
#ifdef USE_JOYSTICK
    struct JS_DATA_TYPE js;
    int fd;

    /* open device file */
    if ((fd = open(joydev, O_RDONLY)) < 0) {
	printf("%s : open error(%d)\n", joydev, errno);
	return NOMOVE;
    }

    st = read(fd, &js, JS_RETURN);
    if (st != JS_RETURN) {
	printf("%s : open error(%d)\n", joydev, errno);
	return NOMOVE;
    }

    if (joyinit == 0) {
        joy_x_max = js.x*2;
	joy_y_max = js.y*2;
	joyinit = 1;
    }

    if (joy_x_max < js.x) joy_x_max = js.x;
    if (joy_y_max < js.y) joy_y_max = js.y;

    if (js.x < joy_x_max/3)   status = LEFT;	/* move left    */
    if (js.x > joy_x_max*2/3) status = RIGHT;	/* move right   */
    if (js.y < joy_y_max/3)   status = UP;	/* move up      */
    if (js.y > joy_y_max*2/3) status = DOWN;	/* move down    */

    /* {^̕D */
    if (js.buttons & 1) status = DUG;		/* left button  */
    if (js.buttons & 2) status = DIG;		/* right button */

    close(fd);
#endif /* USE_JOYSTICK */

    return status;
}

/*
 *   L[Cxg荞
 *
 *   IN  : Ȃ
 *
 *   OUT : ret 
 *           NOMOVE : L[͂Ȃ
 *           UP,DOWN,LEFT,RIGHT,DIG,DUG,START : eL[ꂽB
 *
 */
int get_key()
{
#if 0
    if(XCheckWindowEvent(d, w , KeyPressMask , &e) == False) {
	return NOMOVE;
    }

    switch (XKeycodeToKeysym(d, e.xkey.keycode, 0)) {
        case XK_Up: 
	case XK_KP_8:
	case XK_i:
	    return UP;

        case XK_Down: 
	case XK_KP_2:
	case XK_m:
	    return DOWN;
	    
        case XK_Left: 
	case XK_KP_4:
	case XK_j:
	    return LEFT;
	    
	case XK_Right: 
	case XK_KP_6:
	case XK_l:
	    return RIGHT;
	    
	case XK_q:
	    printf("xalien Ver %s End..\n",VER);
	    exit(0);
    }
    return NOMOVE;
#endif

    int i, mask=0, inkey=0, joykey=0;

    if (joyf) joykey = get_joystick();	/* V1.02 */
      
    while (XPending(d) > 0) {
        mask = 0;
        XNextEvent(d, &e);
	switch (e.type){
	  case Expose:
	    redraw();
	    break;
		
	  case KeyPress:
	  case KeyRelease:
	    mask = 0;
	    switch (XKeycodeToKeysym(d, e.xkey.keycode, 0)) {
	      case XK_Up: 
	      case XK_KP_8:
	      case XK_i:
		mask = (1 << UP);
		break;
		
	      case XK_Down: 
	      case XK_KP_5:
	      case XK_m:
		mask = (1 << DOWN);
		break;

	      case XK_Left: 
	      case XK_KP_4:
	      case XK_j:
		mask = (1 << LEFT);
		break;

	      case XK_Right: 
	      case XK_KP_6:
	      case XK_l:
		mask = (1 << RIGHT);
		break;

	      case XK_x:
		mask = (1 << DIG);
		break;

	      case XK_z:
		mask = (1 << DUG);
		break;

	      case XK_s:
		mask = (1 << START);
		break;

	      case XK_q:
                exit(0);	
	    }
	    if (e.type == KeyPress) {
	        inkey |= mask;
	    } else {
	        inkey &= ~mask;
	    }
	    break;
          }
    }
    if (inkey == 0) {
        if (joyf) return joykey;		/* V1.02 */
    	return NOMOVE;
    }
    for (i = 0 ; i <= START; i++) {
        if (inkey & (1 << i)) {
	    return i;
      	}
    }
    printf("unknown key detected. (0x%02x)\n",inkey);
    return NOMOVE;
}

/*
 *   Redraw display
 *
 */
void redraw()
{
    int i;

    if (stage == -1) { /* V1.04-A */
        /* not start */
        ScDispHighScoreX(d, w, 5, 60);
        PUT_LOGO(90,3,RED);
        return;
    }

    /* on game */
    put_maze(PUTBG+PUTWALL);
    for (i=0; i<PLAYER; i++) {
	if (kv[i].rest) {
            PUT_KEV(kv[i].x, kv[i].y, kv[i].vec);
	}
    }
    kev_rest_disp();				/* g̎c\  */
}

/* V1.04-D start */
#if 0
/*
 *   Get HIGH SCORE
 */
void GetHighScore()
{
    FILE *fp;
    char buf[256];
    char *pt;

    sprintf(buf,"%s/xalien.high",HIGHPATH);
    if (!(fp = fopen(buf,"r"))) {
    	perror("xalien");
    	printf("High Score file (%s) read open error.\n",buf);
    	return;
    }
    fgets(buf, sizeof(buf), fp);
    HI_SCORE = atoi(buf);
    pt = (char *)strchr(buf, ':');
    strcpy(HI_NAME, ++pt);
    if (strlen(HI_NAME)) {
	if (HI_NAME[strlen(HI_NAME)-1] == '\n') {
	    HI_NAME[strlen(HI_NAME)-1] = '\0';
	}
    }
    fclose(fp);
}

/*
 *   Save HIGH SCORE
 */
void SaveHighScore()
{
    FILE *fp;
    char buf[256];
    char *pt;
    struct passwd *pwp = getpwuid(getuid());

    sprintf(buf,"%s/xalien.high",HIGHPATH);
    if (!(fp = fopen(buf,"w"))) {
    	perror("xalien");
    	printf("High Score file (%s) write open error.\n",buf);
    	return;
    }
    if (pwp) {
	strcpy(buf,pwp->pw_gecos);
	pt = (char *)strchr(buf,',');
	if (pt) {
	    *pt = '\0';
	}
        sprintf(HI_NAME,"%s (%s)",buf,pwp->pw_name);
    } else {
        sprintf(HI_NAME,"anonymous (%d)",getuid());
    }
    fprintf(fp,"%d:%s",HI_SCORE,HI_NAME);
    fclose(fp);
}
#endif /* 0 */
/* V1.04-D end */

/*
 *   dump maze map to stdout
 */
int dump_map()
{
	int i, j, c;

	for (j = 0; j<54; j++) {
	    for (i = 0; i<46; i++) {
		if (c = map[i+j*46]) {
		    if (c == KABE) {
		        printf("");
		    } else if (ALIEN0 <= c && c <= ALIEN3) {
		        printf("");
		    } else if (KEVA0 <= c && c <= KEVC3) {
		        printf("");
		    } else {
		        printf("");
		    }
		} else {
		    printf("  ");
		}
	    }
	    printf("\n");
	}
}
